#ifndef PANGULU_CUDA_INTERFACE_H
#define PANGULU_CUDA_INTERFACE_H
#ifdef __cplusplus
extern "C"
{
#endif
#include "pangulu_common.h"
#include "pangulu_cuda.h"
    /*
The function in this file is primarily called kernel,these functions would be called in pangulu_testl.h,pangulu_numerical.h,pangulu_preprocess.h
*/
    void pangulu_cuda_device_init(int rank)
    {
        int gpu_num;
        pangulu_cuda_getDevicenum(&gpu_num);
        int usr_id = pangulu_cuda_setDevice(gpu_num, rank);
        printf("the gpu num is %d the rank is %d use the %d gpu\n", gpu_num, rank, usr_id);
    }

    void pangulu_cuda_malloc_init_pangulu_Smatrix(pangulu_Smatrix *S)
    {
        pangulu_cuda_malloc_int(&(S->CUDA_rowpointer), (S->row) + 1);
        pangulu_cuda_malloc_int(&(S->CUDA_columnindex), S->rowpointer[S->row]);
        pangulu_cuda_malloc_double(&(S->CUDA_value), S->rowpointer[S->row]);
        pangulu_cuda_malloc_int(&(S->CUDA_columnpointer), (S->row) + 1);
        pangulu_cuda_malloc_int(&(S->CUDA_rowindex), S->rowpointer[S->row]);
        pangulu_cuda_malloc_double(&(S->CUDA_value_CSC), S->rowpointer[S->row]);
        pangulu_cuda_malloc_int(&(S->CUDA_CSR_to_CSC_index), S->rowpointer[S->row]);
        pangulu_cuda_malloc_int(&(S->CUDA_CSC_to_CSR_index), S->rowpointer[S->row]);
        pangulu_cuda_malloc_int(&(S->CUDA_bin_rowpointer), BIN_LENGTH);
        pangulu_cuda_malloc_int(&(S->CUDA_bin_rowindex), S->row);
    }

    void pangulu_cuda_memcpy_init_pangulu_Smatrix(pangulu_Smatrix *S)
    {
        pangulu_cuda_memcpy_host_to_device_int(S->CUDA_rowpointer, S->rowpointer, (S->row) + 1);
        pangulu_cuda_memcpy_host_to_device_int(S->CUDA_columnindex, S->columnindex, S->rowpointer[S->row]);
        pangulu_cuda_memcpy_host_to_device_int(S->CUDA_columnpointer, S->columnpointer, (S->row) + 1);
        pangulu_cuda_memcpy_host_to_device_int(S->CUDA_rowindex, S->rowindex, S->rowpointer[S->row]);
        pangulu_cuda_memcpy_host_to_device_int(S->CUDA_CSR_to_CSC_index, S->CSR_to_CSC_index, S->rowpointer[S->row]);
        pangulu_cuda_memcpy_host_to_device_int(S->CUDA_CSC_to_CSR_index, S->CSC_to_CSR_index, S->rowpointer[S->row]);
        pangulu_cuda_memcpy_host_to_device_double(S->CUDA_value, S->value, S->rowpointer[S->row]);
        pangulu_cuda_memcpy_host_to_device_double(S->CUDA_value_CSC, S->value_CSC, S->rowpointer[S->row]);
        pangulu_cuda_memcpy_host_to_device_int(S->CUDA_bin_rowpointer, S->bin_rowpointer, BIN_LENGTH);
        pangulu_cuda_memcpy_host_to_device_int(S->CUDA_bin_rowindex, S->bin_rowindex, S->row);
    }

    void pangulu_cuda_memcpy_value_host_to_device(pangulu_Smatrix *S)
    {
        pangulu_cuda_memcpy_host_to_device_double(S->CUDA_value, S->value, S->rowpointer[S->row]);
    }

    void pangulu_cuda_memcpy_value_host_to_device_CSC(pangulu_Smatrix *S)
    {
        pangulu_cuda_memcpy_host_to_device_double(S->CUDA_value_CSC, S->value_CSC, S->columnpointer[S->row]);
    }

    void pangulu_cuda_memcpy_value_device_to_host(pangulu_Smatrix *S)
    {
        pangulu_cuda_memcpy_device_to_host_double(S->value, S->CUDA_value, S->rowpointer[S->row]);
    }

    void pangulu_cuda_memcpy_value_device_to_host_CSC(pangulu_Smatrix *S)
    {
        pangulu_cuda_memcpy_device_to_host_double(S->value_CSC, S->CUDA_value_CSC, S->columnpointer[S->row]);
    }

    void pangulu_cuda_transport_CSR_to_CSC(pangulu_Smatrix *S)
    {
        pangulu_cuda_transport_kernel(S->columnpointer[S->row],
                                      S->CUDA_value,
                                      S->CUDA_CSR_to_CSC_index,
                                      S->CUDA_value_CSC);
    }

    void pangulu_cuda_transport_CSC_to_CSR(pangulu_Smatrix *S)
    {
        pangulu_cuda_transport_kernel(S->rowpointer[S->row],
                                      S->CUDA_value_CSC,
                                      S->CUDA_CSC_to_CSR_index,
                                      S->CUDA_value);
    }

    void pangulu_sptrsv_cuda_malloc_init(double **spmv_CUDA_x,
                                         double **spmv_CUDA_y,
                                         double **sptrsv_CUDA_x,
                                         double **sptrsv_CUDA_b,
                                         int **d_graphInDegree,
                                         int **d_graphInDegree_backup,
                                         double **d_left_sum,
                                         int **d_id_extractor,
                                         int n)
    {
        pangulu_cuda_malloc_double(spmv_CUDA_x, n);
        pangulu_cuda_malloc_double(spmv_CUDA_y, n);
        pangulu_cuda_malloc_double(sptrsv_CUDA_x, n);
        pangulu_cuda_malloc_double(sptrsv_CUDA_b, n);
        pangulu_cuda_malloc_int(d_graphInDegree, n);
        pangulu_cuda_malloc_int(d_graphInDegree_backup, n);
        pangulu_cuda_malloc_double(d_left_sum, n);
        pangulu_cuda_malloc_int(d_id_extractor, n);
    }

    void pangulu_sptrsv_cuda_malloc_free(double **spmv_CUDA_x,
                                         double **spmv_CUDA_y,
                                         double **sptrsv_CUDA_x,
                                         double **sptrsv_CUDA_b,
                                         int **d_graphInDegree,
                                         int **d_graphInDegree_backup,
                                         double **d_left_sum,
                                         int **d_id_extractor,
                                         int n)
    {
        *spmv_CUDA_x = pangulu_cuda_free_double(*spmv_CUDA_x);
        *spmv_CUDA_y = pangulu_cuda_free_double(*spmv_CUDA_y);
        *sptrsv_CUDA_x = pangulu_cuda_free_double(*sptrsv_CUDA_x);
        *sptrsv_CUDA_b = pangulu_cuda_free_double(*sptrsv_CUDA_b);
        *d_left_sum = pangulu_cuda_free_double(*d_left_sum);
        *d_graphInDegree = pangulu_cuda_free_int(*d_graphInDegree);
        *d_graphInDegree_backup = pangulu_cuda_free_int(*d_graphInDegree_backup);
        *d_id_extractor = pangulu_cuda_free_int(*d_id_extractor);
    }

#ifdef __cplusplus
}
#endif

#endif