#ifndef PANGULU_PREPROCESS_H
#define PANGULU_PREPROCESS_H

#ifdef __cplusplus
extern "C"
{
#endif
#include "pangulu_common.h"
#include "pangulu_cuda_interface.h"
    /*An impoortant function is pangulu_preprocess(int N,
                            int rank,
                            int P,
                            int Q,
                            int NB,
                            int block_length,
                            int rank_row_length,
                            int rank_col_length,
                            pangulu_Smatrix *symbolic_Smatrix,
                            int **mapper_Big_pangulu_Smatrix,
                            int **Big_pangulu_Smatrix_rowpointer,
                            int **Big_pangulu_Smatrix_columnindex,
                            pangulu_Smatrix ***Big_pangulu_Smatrix_value,
                            int **L_pangulu_Smatrix_columnpointer,
                            int **L_pangulu_Smatrix_rowindex,
                            pangulu_Smatrix ***L_pangulu_Smatrix_value,
                            int **U_pangulu_Smatrix_rowpointer,
                            int **U_pangulu_Smatrix_columnindex,
                            pangulu_Smatrix ***U_pangulu_Smatrix_value){ }
It's parameters are explained as follows: 
N is the size of the original matrix;
rank is this process number;
P is the number of process grid rows;
Q is the number of process grid columns;
NB is the size of the sub-matrix block;
block_length is the number of rows of the matrix after division;
rank_row_length is the row length of the current process matrix.
rank_col_length is the col length of the current process matrix.
symbolic_Smatrix is  the matrix that is broken down by symbols;
mapper_Big_pangulu_Smatrix is the mapping relationship of matrix A.
Big_pangulu_Smatrix_rowpointer is the row offset of the A matrix;
Big_pangulu_Smatrix_columnindex is the subscript of the non-zero elements column of A matrix;
Big_pangulu_Smatrix_value is element value of A matrix;
L_pangulu_Smatrix_columnpointer is the column offset of the L matrix;
L_pangulu_Smatrix_rowindex is the subscript of the non-zero elements row of L matrix;
L_pangulu_Smatrix_value is element value of L matrix;
U_pangulu_Smatrix_rowpointer is the row offset of the U matrix;
U_pangulu_Smatrix_columnindex is the subscript of the non-zero elements column of U matrix;
U_pangulu_Smatrix_value s element value of U matrix.
*/
    void pangulu_get_pangulu_Smatrix_to_U(pangulu_Smatrix *S,
                                          pangulu_Smatrix *U,
                                          int NB)
    {
        int *rowpointer = (int *)malloc(sizeof(int) * (NB + 1));
        rowpointer[0] = 0;
        for (int i = 0; i < NB; i++)
        {
            rowpointer[i + 1] = rowpointer[i];
            for (int j = S->rowpointer[i]; j < S->rowpointer[i + 1]; j++)
            {
                int col = S->columnindex[j];
                if (i <= col)
                {
                    rowpointer[i + 1]++;
                }
            }
        }
        int *columnindex = (int *)malloc(sizeof(int) * rowpointer[NB]);
        calculate_type *value = (calculate_type *)malloc(sizeof(calculate_type) * rowpointer[NB]);
        for (int i = 0; i < rowpointer[NB]; i++)
        {
            value[i] = 0.0;
        }
        int nzz = 0;
        for (int i = 0; i < NB; i++)
        {
            for (int j = S->rowpointer[i]; j < S->rowpointer[i + 1]; j++)
            {
                int col = S->columnindex[j];
                if (i <= col)
                {
                    columnindex[nzz] = col;
                    value[nzz] = S->value[j];
                    nzz++;
                }
            }
        }
        U->rowpointer = rowpointer;
        U->columnindex = columnindex;
        U->value = value;
        U->row = NB;
        U->column = NB;
        return;
    }

    void pangulu_get_pangulu_Smatrix_to_L(pangulu_Smatrix *S,
                                          pangulu_Smatrix *L,
                                          int NB)
    {
        int *rowpointer = (int *)malloc(sizeof(int) * (NB + 1));
        rowpointer[0] = 0;
        for (int i = 0; i < NB; i++)
        {
            rowpointer[i + 1] = rowpointer[i];
            for (int j = S->rowpointer[i]; j < S->rowpointer[i + 1]; j++)
            {
                int col = S->columnindex[j];
                if (i >= col)
                {
                    rowpointer[i + 1]++;
                }
            }
        }
        int *columnindex = (int *)malloc(sizeof(int) * rowpointer[NB]);
        calculate_type *value = (calculate_type *)malloc(sizeof(calculate_type) * rowpointer[NB]);
        for (int i = 0; i < rowpointer[NB]; i++)
        {
            value[i] = 0.0;
        }
        int nzz = 0;
        for (int i = 0; i < NB; i++)
        {
            for (int j = S->rowpointer[i]; j < S->rowpointer[i + 1]; j++)
            {
                int col = S->columnindex[j];
                if (i >= col)
                {
                    columnindex[nzz] = col;
                    value[nzz] = S->value[j];
                    nzz++;
                }
            }
        }
        L->rowpointer = rowpointer;
        L->columnindex = columnindex;
        L->value = value;
        L->row = NB;
        L->column = NB;
        return;
    }

    void pangulu_send_panglu_Smatrix_row_columnindex(int rank,
                                                     int level,
                                                     int block_length,
                                                     int P, int Q, int NB, int signal,
                                                     pangulu_Smatrix *S)
    {
        int send_id = rank;
        for (int i = 0; i < Q - 1; i++)
        {
            if ((send_id + 1) % Q == 0)
            {
                send_id -= Q;
            }
            send_id += 1;
            pangulu_send_pangulu_Smatrix_columnindex(S, send_id, signal, NB);
        }
    }

    void pangulu_send_panglu_Smatrix_col_columnindex(int rank, int level, int block_length,
                                                     int P, int Q, int NB, int signal,
                                                     pangulu_Smatrix *S)
    {
        int send_id = rank;
        for (int i = 0; i < P - 1; i++)
        {
            send_id += Q;
            if (send_id >= P * Q)
            {
                send_id = send_id % Q;
            }
            pangulu_send_pangulu_Smatrix_columnindex(S, send_id, signal, NB);
        }
    }

    void pangulu_divide_Big_pangulu_Smatrix_to_A_L_U(int N,
                                                     int rank,
                                                     int P,
                                                     int Q,
                                                     int NB,
                                                     int block_length,
                                                     int rank_row_length,
                                                     int rank_col_length,
                                                     pangulu_Smatrix *symbolic_Smatrix,
                                                     int **mapper_Big_pangulu_Smatrix,
                                                     int **Big_pangulu_Smatrix_rowpointer,
                                                     int **Big_pangulu_Smatrix_columnindex,
                                                     pangulu_Smatrix ***Big_pangulu_Smatrix_value,
                                                     int **L_pangulu_Smatrix_columnpointer,
                                                     int **L_pangulu_Smatrix_rowindex,
                                                     pangulu_Smatrix ***L_pangulu_Smatrix_value,
                                                     int **U_pangulu_Smatrix_rowpointer,
                                                     int **U_pangulu_Smatrix_columnindex,
                                                     pangulu_Smatrix ***U_pangulu_Smatrix_value)
    {
        int A_nnz_rowpointer_num;
        int *block_Smatrix_nnzA_num = (int *)malloc(sizeof(int) * block_length * block_length);
        int *block_Smatrix_non_zero_num;
        int *block_Smatrix_non_zero_vector_L;
        int *block_Smatrix_non_zero_vector_U;
        if (rank == 0)
        {
            for (int i = 0; i < block_length * block_length; i++)
            {
                block_Smatrix_nnzA_num[i] = 0;
            }
            A_nnz_rowpointer_num = 0;
            for (int i = 0; i < N; i++)
            {
                for (int j = symbolic_Smatrix->rowpointer[i]; j < symbolic_Smatrix->rowpointer[i + 1]; j++)
                {
                    int block_row = i / NB;
                    int block_col = (symbolic_Smatrix->columnindex[j]) / NB;
                    block_Smatrix_nnzA_num[block_row * block_length + block_col]++;
                }
            }
            for (int i = 0; i < block_length * block_length; i++)
            {
                if (block_Smatrix_nnzA_num[i] != 0)
                {
                    block_Smatrix_nnzA_num[i] = A_nnz_rowpointer_num;
                    A_nnz_rowpointer_num++;
                }
                else
                {
                    block_Smatrix_nnzA_num[i] = -1;
                }
            }
            block_Smatrix_non_zero_num = (int *)malloc(sizeof(int) * NB * A_nnz_rowpointer_num);
            block_Smatrix_non_zero_vector_L = (int *)malloc(sizeof(int) * NB * block_length);
            block_Smatrix_non_zero_vector_U = (int *)malloc(sizeof(int) * NB * block_length);
            for (int i = 0; i < A_nnz_rowpointer_num * NB; i++)
            {
                block_Smatrix_non_zero_num[i] = 0;
            }
            for (int i = 0; i < block_length * NB; i++)
            {
                block_Smatrix_non_zero_vector_L[i] = 0;
            }
            for (int i = 0; i < block_length * NB; i++)
            {
                block_Smatrix_non_zero_vector_U[i] = 0;
            }
            for (int i = 0; i < N; i++)
            {
                for (int j = symbolic_Smatrix->rowpointer[i]; j < symbolic_Smatrix->rowpointer[i + 1]; j++)
                {
                    int block_row = i / NB;
                    int block_col = (symbolic_Smatrix->columnindex[j]) / NB;

                    block_Smatrix_non_zero_num[block_Smatrix_nnzA_num[block_row * block_length + block_col] * NB + i % NB]++;
                    if (block_row == block_col)
                    {
                        if (i >= symbolic_Smatrix->columnindex[j])
                        {
                            block_Smatrix_non_zero_vector_L[block_row * NB + i % NB]++;
                        }
                        if (i <= symbolic_Smatrix->columnindex[j])
                        {
                            block_Smatrix_non_zero_vector_U[block_col * NB + i % NB]++;
                        }
                    }
                }
            }
            for (int i = 0; i < A_nnz_rowpointer_num; i++)
            {
                int *now_Smatrix = block_Smatrix_non_zero_num + i * NB;
                for (int j = 1; j < NB; j++)
                {
                    now_Smatrix[j] += now_Smatrix[j - 1];
                }
            }
            for (int i = 0; i < block_length; i++)
            {
                int *now_Smatrix = block_Smatrix_non_zero_vector_L + i * NB;
                for (int j = 1; j < NB; j++)
                {
                    now_Smatrix[j] += now_Smatrix[j - 1];
                }
            }
            for (int i = 0; i < block_length; i++)
            {
                int *now_Smatrix = block_Smatrix_non_zero_vector_U + i * NB;
                for (int j = 1; j < NB; j++)
                {
                    now_Smatrix[j] += now_Smatrix[j - 1];
                }
            }
        }
        pangulu_Bcast_vector(&A_nnz_rowpointer_num, 1, 0);
        pangulu_Bcast_vector(block_Smatrix_nnzA_num, block_length * block_length, 0);
        if (rank != 0)
        {
            block_Smatrix_non_zero_num = (int *)malloc(sizeof(int) * NB * A_nnz_rowpointer_num);
            block_Smatrix_non_zero_vector_L = (int *)malloc(sizeof(int) * NB * block_length);
            block_Smatrix_non_zero_vector_U = (int *)malloc(sizeof(int) * NB * block_length);
        }
        pangulu_Bcast_vector(block_Smatrix_non_zero_num, NB * A_nnz_rowpointer_num, 0);
        pangulu_Bcast_vector(block_Smatrix_non_zero_vector_L, NB * block_length, 0);
        pangulu_Bcast_vector(block_Smatrix_non_zero_vector_U, NB * block_length, 0);
        int *Big_Smatrix_rowpointer = malloc(sizeof(int) * (rank_row_length + 1));
        int nnzA = 0;
        for (int i = 0; i < (rank_row_length + 1); i++)
        {
            Big_Smatrix_rowpointer[i] = 0;
        }
        int now_length = 0;
        for (int i = rank / Q; i < block_length; i += P)
        {
            Big_Smatrix_rowpointer[now_length + 1] = Big_Smatrix_rowpointer[now_length];
            for (int j = rank % Q; j < block_length; j += Q)
            {
                int find_index = block_Smatrix_nnzA_num[i * block_length + j];
                if (find_index != -1)
                {
                    int *now_block_address = block_Smatrix_non_zero_num + NB * find_index;
                    Big_Smatrix_rowpointer[now_length + 1]++;
                    nnzA += (now_block_address[NB - 1]);
                }
            }
            now_length++;
        }
        int *Big_Smatrix_columnindex = (int *)malloc(sizeof(int) * Big_Smatrix_rowpointer[rank_row_length]);
        pangulu_Smatrix **Big_Smatrix_value = (pangulu_Smatrix **)malloc(sizeof(pangulu_Smatrix *) * Big_Smatrix_rowpointer[rank_row_length]);
        int index = 0;
        for (int i = rank / Q; i < block_length; i += P)
        {
            for (int j = rank % Q; j < block_length; j += Q)
            {
                int find_index = block_Smatrix_nnzA_num[i * block_length + j];
                if (find_index != -1)
                {
                    int *now_block_address = block_Smatrix_non_zero_num + NB * find_index;
                    Big_Smatrix_columnindex[index] = j;
                    Big_Smatrix_value[index] = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                    pangulu_init_pangulu_Smatrix(Big_Smatrix_value[index]);
                    pangulu_malloc_pangulu_Smatrix(Big_Smatrix_value[index], NB, now_block_address);
                    index++;
                }
            }
        }
        int *save_Smatrix_columnindex;
        calculate_type *save_Smatrix_value;
        if (rank == 0)
        {
            save_Smatrix_columnindex = (int *)malloc(sizeof(int) * symbolic_Smatrix->rowpointer[N]);
            save_Smatrix_value = (calculate_type *)malloc(sizeof(calculate_type) * symbolic_Smatrix->rowpointer[N]);
            int *rank_nnzA = (int *)malloc(sizeof(int) * (P * Q + 1));
            int *save_rank_nnzA = (int *)malloc(sizeof(int) * (P * Q + 1));
            for (int i = 0; i <= P * Q; i++)
            {
                rank_nnzA[i] = 0;
            }
            for (int i = 0; i < block_length; i++)
            {
                for (int j = 0; j < block_length; j++)
                {
                    int find_index = block_Smatrix_nnzA_num[i * block_length + j];
                    if (find_index != -1)
                    {
                        int now_nnz = block_Smatrix_non_zero_num[find_index * NB + NB - 1];
                        rank_nnzA[i % P * Q + j % Q + 1] += now_nnz;
                    }
                }
            }
            for (int i = 0; i < P * Q; i++)
            {
                rank_nnzA[i + 1] += rank_nnzA[i];
            }
            for (int i = 0; i <= P * Q; i++)
            {
                save_rank_nnzA[i] = rank_nnzA[i];
            }
            for (int i = 0; i < N; i++)
            {
                for (int j = symbolic_Smatrix->rowpointer[i]; j < symbolic_Smatrix->rowpointer[i + 1]; j++)
                {
                    int block_row = i / NB;
                    int block_col = (symbolic_Smatrix->columnindex[j]) / NB;
                    int index = block_row % P * Q + block_col % Q;
                    save_Smatrix_columnindex[rank_nnzA[index]] = symbolic_Smatrix->columnindex[j] % NB;
                    save_Smatrix_value[rank_nnzA[index]] = symbolic_Smatrix->value[j];
                    rank_nnzA[index]++;
                }
            }
            for (int i = 1; i < P * Q; i++)
            {
                pangulu_send_vector_int(save_Smatrix_columnindex + save_rank_nnzA[i],
                                        save_rank_nnzA[i + 1] - save_rank_nnzA[i], i, i);
            }
            for (int i = 1; i < P * Q; i++)
            {
                pangulu_send_vector_double(save_Smatrix_value + save_rank_nnzA[i],
                                           save_rank_nnzA[i + 1] - save_rank_nnzA[i], i, i);
            }
            free(rank_nnzA);
            free(save_rank_nnzA);
        }
        else
        {
            save_Smatrix_columnindex = (int *)malloc(sizeof(int) * nnzA);
            save_Smatrix_value = (calculate_type *)malloc(sizeof(calculate_type) * nnzA);
            pangulu_recv_vector_int(save_Smatrix_columnindex, nnzA, 0, rank);
            pangulu_recv_vector_double(save_Smatrix_value, nnzA, 0, rank);
        }
        int get_nnz = 0;
        for (int i = 0; i < rank_row_length; i++)
        {
            for (int j = 0; j < NB; j++)
            {
                for (int k = Big_Smatrix_rowpointer[i]; k < Big_Smatrix_rowpointer[i + 1]; k++)
                {
                    pangulu_Smatrix *now_Smatrix = Big_Smatrix_value[k];
                    for (int m = now_Smatrix->rowpointer[j]; m < now_Smatrix->rowpointer[j + 1]; m++)
                    {
                        now_Smatrix->columnindex[m] = save_Smatrix_columnindex[get_nnz];
                        now_Smatrix->value[m] = save_Smatrix_value[get_nnz];
                        get_nnz++;
                    }
                }
            }
        }
        free(save_Smatrix_columnindex);
        free(save_Smatrix_value);
        int *mapper_index = (int *)malloc(sizeof(int) * rank_row_length * rank_col_length);
        for (int i = 0; i < rank_row_length; i++)
        {
            for (int j = 0; j < rank_col_length; j++)
            {
                mapper_index[i * rank_col_length + j] = -1;
            }
        }
        for (int i = 0; i < rank_row_length; i++)
        {
            for (int j = Big_Smatrix_rowpointer[i]; j < Big_Smatrix_rowpointer[i + 1]; j++)
            {
                mapper_index[i * rank_col_length + Big_Smatrix_columnindex[j] / Q] = j;
            }
        }
        int *U_rowpointer = (int *)malloc(sizeof(int) * (block_length + 1));
        int *L_rowpointer = (int *)malloc(sizeof(int) * (block_length + 1));
        U_rowpointer[0] = 0;
        L_rowpointer[0] = 0;
        int offset_row = rank / Q;
        int offset_col = rank % Q;
        for (int level = 0; level < block_length; level++)
        {
            U_rowpointer[level + 1] = U_rowpointer[level];
            L_rowpointer[level + 1] = L_rowpointer[level];
            for (int i = offset_col + level; i < block_length; i += Q)
            {
                int find_index = block_Smatrix_nnzA_num[level * block_length + i];
                if (find_index != -1)
                {
                    U_rowpointer[level + 1]++;
                }
            }
            for (int i = offset_row + level; i < block_length; i += P)
            {
                int find_index = block_Smatrix_nnzA_num[i * block_length + level];
                if (find_index != -1)
                {
                    L_rowpointer[level + 1]++;
                }
            }
            offset_row = pangulu_exchange_PQ(offset_row, P);
            offset_col = pangulu_exchange_PQ(offset_col, Q);
        }
        if (rank == -1)
        {
            for (int i = 0; i < Big_Smatrix_rowpointer[rank_row_length]; i++)
            {
                pangulu_display_pangulu_Smatrix(Big_Smatrix_value[i]);
            }
        }
        int U_Smatrix_nzz = 0;
        int L_Smatrix_nzz = 0;
        int *U_columnindex = (int *)malloc(sizeof(int) * U_rowpointer[block_length]);
        pangulu_Smatrix **U_value = (pangulu_Smatrix **)malloc(sizeof(pangulu_Smatrix *) * U_rowpointer[block_length]);
        int *L_columnindex = (int *)malloc(sizeof(int) * L_rowpointer[block_length]);
        pangulu_Smatrix **L_value = (pangulu_Smatrix **)malloc(sizeof(pangulu_Smatrix *) * L_rowpointer[block_length]);
        offset_row = rank / Q;
        offset_col = rank % Q;
        for (int level = 0; level < block_length; level++)
        {
            if (offset_row == 0 && offset_col == 0)
            {
                int first_index = pangulu_mapper_to_Big_pangulu_Smatrix(level, level, mapper_index,
                                                                        rank_col_length, P, Q);
                pangulu_Smatrix *first_U = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                pangulu_init_pangulu_Smatrix(first_U);
                pangulu_get_pangulu_Smatrix_to_U(Big_Smatrix_value[first_index], first_U, NB);
                U_value[U_Smatrix_nzz] = first_U;
                U_columnindex[U_Smatrix_nzz] = level;
                //send
                pangulu_send_panglu_Smatrix_col_columnindex(rank, level, block_length, P, Q, NB,
                                                            level * block_length + level,
                                                            first_U);
                U_Smatrix_nzz++;
                pangulu_Smatrix *first_L = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                pangulu_init_pangulu_Smatrix(first_L);
                pangulu_get_pangulu_Smatrix_to_L(Big_Smatrix_value[first_index], first_L, NB);
                L_value[L_Smatrix_nzz] = first_L;
                L_columnindex[L_Smatrix_nzz] = level;
                //send
                pangulu_send_panglu_Smatrix_row_columnindex(rank, level, block_length, P, Q, NB,
                                                            level * block_length + level,
                                                            first_L);
                L_Smatrix_nzz++;
                for (int i = Q + level; i < block_length; i += Q)
                {
                    int find_index = block_Smatrix_nnzA_num[level * block_length + i];
                    if (find_index != -1)
                    {
                        //index
                        int col_index = pangulu_mapper_to_Big_pangulu_Smatrix(level, i, mapper_index,
                                                                              rank_col_length, P, Q);
                        pangulu_Smatrix *U = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                        pangulu_init_pangulu_Smatrix(U);
                        pangulu_copy_pangulu_Smatrix_to_pangulu_Smatrix(Big_Smatrix_value[col_index], U, NB);
                        U_value[U_Smatrix_nzz] = U;
                        U_columnindex[U_Smatrix_nzz] = i;
                        //send
                        pangulu_send_panglu_Smatrix_col_columnindex(rank, level, block_length, P, Q, NB,
                                                                    level * block_length + i,
                                                                    U);
                        U_Smatrix_nzz++;
                    }
                }
                for (int i = P + level; i < block_length; i += P)
                {
                    int find_index = block_Smatrix_nnzA_num[i * block_length + level];
                    if (find_index != -1)
                    {
                        //index
                        int row_index = pangulu_mapper_to_Big_pangulu_Smatrix(i, level, mapper_index,
                                                                              rank_col_length, P, Q);
                        pangulu_Smatrix *L = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                        pangulu_init_pangulu_Smatrix(L);
                        pangulu_copy_pangulu_Smatrix_to_pangulu_Smatrix(Big_Smatrix_value[row_index], L, NB);
                        L_value[L_Smatrix_nzz] = L;
                        L_columnindex[L_Smatrix_nzz] = i;
                        //send
                        pangulu_send_panglu_Smatrix_row_columnindex(rank, level, block_length, P, Q, NB,
                                                                    i * block_length + level,
                                                                    L);
                        L_Smatrix_nzz++;
                    }
                }
            }
            if (offset_col != 0 && offset_row == 0)
            {
                pangulu_Smatrix *first_L = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                pangulu_init_pangulu_Smatrix(first_L);
                pangulu_malloc_pangulu_Smatrix(first_L, NB, block_Smatrix_non_zero_vector_L + NB * level);
                //receive
                int receive_id = rank - offset_col;
                if (receive_id < (rank - rank % Q))
                {
                    receive_id += Q;
                }
                pangulu_recv_pangulu_Smatrix_columnindex(first_L, receive_id, level * block_length + level, NB);
                L_value[L_Smatrix_nzz] = first_L;
                L_columnindex[L_Smatrix_nzz] = level;
                L_Smatrix_nzz++;
                for (int i = level + offset_col; i < block_length; i += Q)
                {
                    int find_index = block_Smatrix_nnzA_num[level * block_length + i];
                    if (find_index != -1)
                    {
                        //index
                        int col_index = pangulu_mapper_to_Big_pangulu_Smatrix(level, i, mapper_index,
                                                                              rank_col_length, P, Q);
                        pangulu_Smatrix *U = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                        pangulu_init_pangulu_Smatrix(U);
                        pangulu_copy_pangulu_Smatrix_to_pangulu_Smatrix(Big_Smatrix_value[col_index], U, NB);
                        U_value[U_Smatrix_nzz] = U;
                        U_columnindex[U_Smatrix_nzz] = i;
                        //send
                        pangulu_send_panglu_Smatrix_col_columnindex(rank, level, block_length, P, Q, NB,
                                                                    level * block_length + i,
                                                                    U);
                        U_Smatrix_nzz++;
                    }
                }
                for (int i = level + P; i < block_length; i += P)
                {
                    int find_index = block_Smatrix_nnzA_num[i * block_length + level];
                    if (find_index != -1)
                    {
                        int *now_block_address = block_Smatrix_non_zero_num + NB * find_index;
                        pangulu_Smatrix *L = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                        pangulu_init_pangulu_Smatrix(L);
                        //malloc
                        pangulu_malloc_pangulu_Smatrix(L, NB, now_block_address);
                        L_value[L_Smatrix_nzz] = L;
                        L_columnindex[L_Smatrix_nzz] = i;
                        //receive
                        pangulu_recv_pangulu_Smatrix_columnindex(L, receive_id, i * block_length + level, NB);
                        L_Smatrix_nzz++;
                    }
                }
            }
            if (offset_row != 0 && offset_col == 0)
            {
                pangulu_Smatrix *first_U = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                pangulu_init_pangulu_Smatrix(first_U);
                pangulu_malloc_pangulu_Smatrix(first_U, NB, block_Smatrix_non_zero_vector_U + NB * level);
                //reiceive
                int receive_id = rank - offset_row * Q;
                if (receive_id < (rank % Q))
                {
                    receive_id += P * Q;
                }
                pangulu_recv_pangulu_Smatrix_columnindex(first_U, receive_id, level * block_length + level, NB);
                U_value[U_Smatrix_nzz] = first_U;
                U_columnindex[U_Smatrix_nzz] = level;
                U_Smatrix_nzz++;
                for (int i = level + offset_row; i < block_length; i += P)
                {
                    int find_index = block_Smatrix_nnzA_num[i * block_length + level];
                    if (find_index != -1)
                    {
                        //index
                        int row_index = pangulu_mapper_to_Big_pangulu_Smatrix(i, level, mapper_index,
                                                                              rank_col_length, P, Q);
                        pangulu_Smatrix *L = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                        pangulu_init_pangulu_Smatrix(L);
                        pangulu_copy_pangulu_Smatrix_to_pangulu_Smatrix(Big_Smatrix_value[row_index], L, NB);
                        L_value[L_Smatrix_nzz] = L;
                        L_columnindex[L_Smatrix_nzz] = i;
                        //send
                        pangulu_send_panglu_Smatrix_row_columnindex(rank, level, block_length, P, Q, NB,
                                                                    i * block_length + level,
                                                                    L);
                        L_Smatrix_nzz++;
                    }
                }
                for (int i = level + Q; i < block_length; i += Q)
                {
                    int find_index = block_Smatrix_nnzA_num[level * block_length + i];
                    if (find_index != -1)
                    {
                        int *now_block_address = block_Smatrix_non_zero_num + NB * find_index;
                        pangulu_Smatrix *U = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                        pangulu_init_pangulu_Smatrix(U);
                        //malloc
                        U_value[U_Smatrix_nzz] = U;
                        U_columnindex[U_Smatrix_nzz] = i;
                        pangulu_malloc_pangulu_Smatrix(U, NB, now_block_address);
                        //receive
                        pangulu_recv_pangulu_Smatrix_columnindex(U, receive_id, level * block_length + i, NB);
                        U_Smatrix_nzz++;
                    }
                }
            }
            if (offset_row != 0 && offset_col != 0)
            {
                int receive_id;
                receive_id = rank - offset_row * Q;
                if (receive_id < (rank % Q))
                {
                    receive_id += P * Q;
                }
                for (int i = offset_col + level; i < block_length; i += Q)
                {
                    int find_index = block_Smatrix_nnzA_num[level * block_length + i];
                    if (find_index != -1)
                    {
                        int *now_block_address = block_Smatrix_non_zero_num + NB * find_index;
                        pangulu_Smatrix *U = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                        pangulu_init_pangulu_Smatrix(U);
                        //malloc
                        U_value[U_Smatrix_nzz] = U;
                        U_columnindex[U_Smatrix_nzz] = i;
                        pangulu_malloc_pangulu_Smatrix(U, NB, now_block_address);
                        //receive
                        pangulu_recv_pangulu_Smatrix_columnindex(U, receive_id, level * block_length + i, NB);
                        U_Smatrix_nzz++;
                    }
                }
                receive_id = rank - offset_col;
                if (receive_id < (rank - rank % Q))
                {
                    receive_id += Q;
                }
                for (int i = offset_row + level; i < block_length; i += P)
                {
                    int find_index = block_Smatrix_nnzA_num[i * block_length + level];
                    if (find_index != -1)
                    {
                        int *now_block_address = block_Smatrix_non_zero_num + NB * find_index;
                        pangulu_Smatrix *L = (pangulu_Smatrix *)malloc(sizeof(pangulu_Smatrix));
                        pangulu_init_pangulu_Smatrix(L);
                        //malloc
                        pangulu_malloc_pangulu_Smatrix(L, NB, now_block_address);
                        L_value[L_Smatrix_nzz] = L;
                        L_columnindex[L_Smatrix_nzz] = i;
                        //receive
                        pangulu_recv_pangulu_Smatrix_columnindex(L, receive_id, i * block_length + level, NB);
                        L_Smatrix_nzz++;
                    }
                }
            }
            offset_row = pangulu_exchange_PQ(offset_row, P);
            offset_col = pangulu_exchange_PQ(offset_col, Q);
        }
        free(block_Smatrix_nnzA_num);
        free(block_Smatrix_non_zero_num);
        free(block_Smatrix_non_zero_vector_L);
        free(block_Smatrix_non_zero_vector_U);
        MPI_Barrier(MPI_COMM_WORLD);
        for (int i = 0; i < Big_Smatrix_rowpointer[rank_row_length]; i++)
        {
            pangulu_pangulu_Smatrix_init_malloc(Big_Smatrix_value[i]);
        }
        for (int i = 0; i < L_rowpointer[block_length]; i++)
        {
            for (int j = 0; j < L_value[i]->rowpointer[NB]; j++)
            {
                L_value[i]->value[j] = 0;
            }
            pangulu_pangulu_Smatrix_init_malloc(L_value[i]);
        }
        for (int i = 0; i < U_rowpointer[block_length]; i++)
        {
            for (int j = 0; j < U_value[i]->rowpointer[NB]; j++)
            {
                U_value[i]->value[j] = 0;
            }
            pangulu_pangulu_Smatrix_init_malloc(U_value[i]);
            pangulu_cuda_malloc_int(&(U_value[i]->CUDA_nnzU), U_value[i]->row);
            int *nnzU = (int *)malloc(sizeof(int) * (U_value[i]->row));
            for (int j = 0; j < U_value[i]->row; j++)
            {
                nnzU[j] = U_value[i]->columnpointer[j + 1] - U_value[i]->columnpointer[j] - 1;
            }
            pangulu_cuda_memcpy_host_to_device_int(U_value[i]->CUDA_nnzU, nnzU, U_value[i]->row);
            U_value[i]->nnzU = nnzU;
        }
        MPI_Barrier(MPI_COMM_WORLD);
        *mapper_Big_pangulu_Smatrix = mapper_index;
        *Big_pangulu_Smatrix_rowpointer = Big_Smatrix_rowpointer;
        *Big_pangulu_Smatrix_columnindex = Big_Smatrix_columnindex;
        *Big_pangulu_Smatrix_value = Big_Smatrix_value;
        *L_pangulu_Smatrix_columnpointer = L_rowpointer;
        *L_pangulu_Smatrix_rowindex = L_columnindex;
        *L_pangulu_Smatrix_value = L_value;
        *U_pangulu_Smatrix_rowpointer = U_rowpointer;
        *U_pangulu_Smatrix_columnindex = U_columnindex;
        *U_pangulu_Smatrix_value = U_value;
        return;
    }
    void pangulu_preprocess(int N,
                            int rank,
                            int P,
                            int Q,
                            int NB,
                            int block_length,
                            int rank_row_length,
                            int rank_col_length,
                            pangulu_Smatrix *symbolic_Smatrix,
                            int **mapper_Big_pangulu_Smatrix,
                            int **Big_pangulu_Smatrix_rowpointer,
                            int **Big_pangulu_Smatrix_columnindex,
                            pangulu_Smatrix ***Big_pangulu_Smatrix_value,
                            int **L_pangulu_Smatrix_columnpointer,
                            int **L_pangulu_Smatrix_rowindex,
                            pangulu_Smatrix ***L_pangulu_Smatrix_value,
                            int **U_pangulu_Smatrix_rowpointer,
                            int **U_pangulu_Smatrix_columnindex,
                            pangulu_Smatrix ***U_pangulu_Smatrix_value)
    {
        pangulu_divide_Big_pangulu_Smatrix_to_A_L_U(N,
                                                    rank,
                                                    P,
                                                    Q,
                                                    NB,
                                                    block_length,
                                                    rank_row_length,
                                                    rank_col_length,
                                                    symbolic_Smatrix,
                                                    mapper_Big_pangulu_Smatrix,
                                                    Big_pangulu_Smatrix_rowpointer,
                                                    Big_pangulu_Smatrix_columnindex,
                                                    Big_pangulu_Smatrix_value,
                                                    L_pangulu_Smatrix_columnpointer,
                                                    L_pangulu_Smatrix_rowindex,
                                                    L_pangulu_Smatrix_value,
                                                    U_pangulu_Smatrix_rowpointer,
                                                    U_pangulu_Smatrix_columnindex,
                                                    U_pangulu_Smatrix_value);
    }
#ifdef __cplusplus
}
#endif

#endif