#ifndef PANGULU_TEST_H
#define PANGULU_TEST_H
#ifdef __cplusplus
extern "C"
{
#endif
#include "pangulu_common.h"
#include "pangulu_utils.h"
#include "pangulu_preprocess.h"
#include "pangulu_numerical.h"
#include "pangulu_check.h"
#include "pangulu_cuda_interface.h"
#include "pangulu_sptrsv.h"
    /*An impoortant function is void pangulu_test(pangulu_common *common,
                      pangulu_Smatrix *symbolic_Smatrix){ }
It's parameters are explained as follows: 
common is the some parameters save in panguLU;
symbolic_matrix is Solve the matrix.
*/
    void pangulu_test(pangulu_common *common,
                      pangulu_Smatrix *symbolic_Smatrix)
    {
        int rank = common->rank;
        int P = common->P;
        int Q = common->Q;
        int NB = common->NB;
        MPI_Barrier(MPI_COMM_WORLD);
        pangulu_time_start(common);
        common->N = pangulu_Bcast_N(symbolic_Smatrix->row, 0);
        int N = common->N;
        if (rank == 0)
        {
            printf("N is %d ,NNZ is %d\n", N, symbolic_Smatrix->rowpointer[N]);
            printf("now use four kernels in which device:\n");
            printf("GETRF:GPU\n");

#ifdef CUDA_TSTRF
            printf("TSTRF:GPU\n");
#else
        printf("TSTRF:CPU\n");
#endif

#ifdef CUDA_GESSM
            printf("GESSM:GPU\n");
#else
        printf("GESSM:CPU\n");
#endif

            printf("SSSSM:GPU\n");
        }
        pangulu_cuda_device_init(rank);
        int block_length = pangulu_Calculate_Block(N, NB);
        int rank_row_length = (block_length / P + (((block_length % P) > (rank / Q)) ? 1 : 0));
        int rank_col_length = (block_length / Q + (((block_length % Q) > (rank % Q)) ? 1 : 0));
        int offset_row = rank / Q;
        int offset_col = rank % Q;
        int *mapper_Big_pangulu_Smatrix;
        int *Big_pangulu_Smatrix_rowpointer;
        int *Big_pangulu_Smatrix_columnindex;
        pangulu_Smatrix **Big_pangulu_Smatrix_value;
        int *L_pangulu_Smatrix_columnpointer;
        int *L_pangulu_Smatrix_rowindex;
        pangulu_Smatrix **L_pangulu_Smatrix_value;
        int *U_pangulu_Smatrix_rowpointer;
        int *U_pangulu_Smatrix_columnindex;
        pangulu_Smatrix **U_pangulu_Smatrix_value;
        pangulu_preprocess(N,
                           rank,
                           P,
                           Q,
                           NB,
                           block_length,
                           rank_row_length,
                           rank_col_length,
                           symbolic_Smatrix,
                           &mapper_Big_pangulu_Smatrix,
                           &Big_pangulu_Smatrix_rowpointer,
                           &Big_pangulu_Smatrix_columnindex,
                           &Big_pangulu_Smatrix_value,
                           &L_pangulu_Smatrix_columnpointer,
                           &L_pangulu_Smatrix_rowindex,
                           &L_pangulu_Smatrix_value,
                           &U_pangulu_Smatrix_rowpointer,
                           &U_pangulu_Smatrix_columnindex,
                           &U_pangulu_Smatrix_value);
        MPI_Barrier(MPI_COMM_WORLD);
        pangulu_time_stop(common);
        if (rank == 0)
        {
            printf("the preprocess time is %lfs\n", pangulu_get_spend_time(common));
        }
        pangulu_time_start(common);
        pangulu_numerical(rank,
                          offset_row,
                          offset_col,
                          block_length,
                          P,
                          Q,
                          NB,
                          rank_row_length,
                          rank_col_length,
                          mapper_Big_pangulu_Smatrix,
                          Big_pangulu_Smatrix_value,
                          L_pangulu_Smatrix_columnpointer,
                          L_pangulu_Smatrix_rowindex,
                          L_pangulu_Smatrix_value,
                          U_pangulu_Smatrix_rowpointer,
                          U_pangulu_Smatrix_columnindex,
                          U_pangulu_Smatrix_value);
        MPI_Barrier(MPI_COMM_WORLD);
        pangulu_time_stop(common);
        if (rank == 0)
        {
            printf("the numerical time is %lfs\n", pangulu_get_spend_time(common));
        }
        /*
        this function is check the numerical error.
        pangulu_check(N,
                      rank,
                      P,
                      Q,
                      NB,
                      block_length,
                      symbolic_Smatrix,
                      L_pangulu_Smatrix_columnpointer,
                      L_pangulu_Smatrix_rowindex,
                      L_pangulu_Smatrix_value,
                      U_pangulu_Smatrix_rowpointer,
                      U_pangulu_Smatrix_columnindex,
                      U_pangulu_Smatrix_value);
        */
        pangulu_destory_Big_pangulu_Smatrix(rank_row_length,
                                            &Big_pangulu_Smatrix_rowpointer,
                                            &Big_pangulu_Smatrix_columnindex,
                                            &Big_pangulu_Smatrix_value);
        MPI_Barrier(MPI_COMM_WORLD);
        pangulu_time_start(common);
        pangulu_sptrsv(N,
                       rank,
                       P,
                       Q,
                       NB,
                       block_length,
                       symbolic_Smatrix,
                       L_pangulu_Smatrix_columnpointer,
                       L_pangulu_Smatrix_rowindex,
                       L_pangulu_Smatrix_value,
                       U_pangulu_Smatrix_rowpointer,
                       U_pangulu_Smatrix_columnindex,
                       U_pangulu_Smatrix_value);
        MPI_Barrier(MPI_COMM_WORLD);
        pangulu_time_stop(common);
        if (rank == 0)
        {
            //printf("the sptrsv time is %lfs\n", pangulu_get_spend_time(common));
        }
        pangulu_destory_Big_pangulu_Smatrix(block_length,
                                            &L_pangulu_Smatrix_columnpointer,
                                            &L_pangulu_Smatrix_rowindex,
                                            &L_pangulu_Smatrix_value);
        pangulu_destory_Big_pangulu_Smatrix(block_length,
                                            &U_pangulu_Smatrix_rowpointer,
                                            &U_pangulu_Smatrix_columnindex,
                                            &U_pangulu_Smatrix_value);
        return;
    }
#ifdef __cplusplus
}
#endif

#endif
